import pytest

import copyaid.diff
from copyaid.diff import diffadapt

from os import listdir
from pathlib import Path

CASES_DIR = Path(__file__).parent / "cases"


def test_trivial_diffs():
    assert diffadapt("Whatever\n", [""]) == ["\n"]
    assert diffadapt("Hello\n", ["World"]) == ["World\n"]


def print_operation(rev, orig):
    orig_repr = repr("".join(orig))
    rev_repr = repr("".join(rev))
    if rev == orig:
        print("EQU:", rev_repr)
    elif not rev:
        print("DEL:", orig_repr)
    elif not orig:
        print("INS:", rev_repr)
    else:
        print("ORI:", orig_repr)
        print("REV:", rev_repr)


class OperationsPrinter(copyaid.diff.DiffAdaptor):
    
    def _do_operation(self, tag, rev, orig):
        assert (tag == 'equal') == (rev == orig)
        #print_operation(rev, orig)
        ret = super()._do_operation(tag, rev, orig)
        tokens = repr("".join(ret))
        last_token = repr(self.last_token)
        print("DEBT/OUT/LAST: {}/{}/{}".format(self.line_debt, tokens, last_token))

    def _undo_delete(self, orig):
        print_operation([], orig)
        return super()._undo_delete(orig)

    def _adapt_unrevised(self, rev):
        print_operation(rev, rev)
        return super()._adapt_unrevised(rev)

    def _adapt_revised(self, rev, orig):
        print_operation(rev, orig)
        return super()._adapt_revised(rev, orig)


def print_operations(orig_text, rev_text):
    matcher = copyaid.diff.TokenSequenceMatcher(orig_text)
    matcher.set_alternative(rev_text)
    tokens = OperationsPrinter()
    for tag, rev_chunk, orig_chunk in matcher.operations():
        tokens._do_operation(tag, rev_chunk, orig_chunk)


def read_text_files(subcase_dir):
    ret = dict()
    for path in subcase_dir.iterdir():
        assert path.suffix == ".txt"
        with open(path) as file:
            ret[path.stem] = file.read()
    return ret 


@pytest.mark.parametrize("case", listdir(CASES_DIR / "diff"))
def test_diffadapt(case):
    txt = read_text_files(CASES_DIR / "diff" / case)
    #print_operations(txt["orig"], txt["revised"])
    got = diffadapt(txt["orig"], [txt["revised"]])[0]
    assert got == txt["expected"]


@pytest.mark.parametrize("case", listdir(CASES_DIR / "undo"))
def test_diffadapt_undo(case):
    txt = read_text_files(CASES_DIR / "undo" / case)
    #print_operations(txt["orig"], txt["revised"])
    got = diffadapt(txt["orig"], [txt["revised"]])[0]
    assert got == txt["orig"]



import tomli
from .util import copy_package_dir, read_file_text, resolve_path
from .core import (
    ApiProxy, CopybreakSyntax, CopyEditor, SimpleParser, SourceParserProtocol,
    TrivialParser, WorkFiles, warning
)

# Python Standard Library
import io, subprocess
from pathlib import Path
from typing import Any, Iterable, Optional


class Task:
    def __init__(self, ed: CopyEditor, react_cmds: list[str]):
        self._editor = ed
        self._react = react_cmds

    @property
    def can_request(self) -> bool:
        return bool(self._editor) and self._editor.has_instructions

    def request(self, work: WorkFiles) -> None:
        assert self.can_request
        self._editor.revise(work)

    def react(self, work: WorkFiles) -> int:
        ret = 0
        if self._react:
            found_revs = [str(p) for p in work.revisions()]
            if found_revs:
                args = [str(work.src)] + found_revs
                for cmd in self._react:
                    proc = subprocess.run([cmd] + args, shell=True)
                    ret = proc.returncode
                    if ret:
                        break
        return ret


class PackageConfig:
    """
    Class for handling base configuration from package.
    """

    CONFIG_FILENAME = "copyaid.toml"

    def __init__(self, local_dir: Path):
        copy_package_dir("config", local_dir)
        config_file = local_dir / PackageConfig.CONFIG_FILENAME
        with open(config_file, "rb") as file:
            data = tomli.load(file)
        self._formats = data.get("formats", {})
        self._commands = data.get("commands", {})
        self._tasks: dict[str, Any] = dict()
        self._add_task_data(local_dir, data.get("tasks", {}))

    def _add_task_data(self, config_dir: Path, tasks: dict[str, Any]) -> None:
        for key, task in tasks.items():
            self._tasks[key] = dict(
                request=resolve_path(config_dir, task.get("request")),
                react=task.get("react"),
            )

    def _get_parsers(self) -> list[SourceParserProtocol]:
        ret = list()
        for fname, f in self._formats.items():
            if "copybreak" not in f:
                raise SyntaxError(f"Format '{fname}' table missing 'copybreak' key")
            ret.append(SimpleParser.from_POD(f))
        if not ret:
            warning("No file formats specified in config file.")
        return ret + [TrivialParser()]

    def _react_as_commands(self, react: Any) -> list[str]:
        ret = list()
        if react is None:
            react = []
        elif isinstance(react, str):
            react = [react]
        for r in react:
            cmd = self._commands.get(r)
            if cmd is None:
                msg = f"Command '{r}' not found in configuration"
                raise SyntaxError(msg)
            ret.append(cmd)
        return ret


class Config(PackageConfig):
    """
    Class for handling user config files (usually `copyaid.toml`).
    """
    def __init__(self, tmp_dir: Path, config_file: Path):
        super().__init__(tmp_dir)
        if config_file.exists():
            with open(config_file, "rb") as file:
                data = tomli.load(file)
        else:
            data = {}
        self._formats.update(data.get("formats", {}))
        self._commands.update(data.get("commands", {}))
        self._add_task_data(config_file.parent, data.get("tasks", {}))
        key_path = data.get("openai_api_key_file")
        self.api_key = read_file_text(resolve_path(config_file.parent, key_path))
        self.log_format = data.get("log_format")

    @property
    def task_names(self) -> Iterable[str]:
        return self._tasks.keys()

    def get_task(self, task_name: str, log_path: Path) -> Task:
        task = self._tasks.get(task_name)
        if task is None:
            raise ValueError(f"Invalid task name {task_name}.")
        if "clean" in task:
            warning("Configuration setting 'clean' has been deprecated.")
        api = ApiProxy(self.api_key, log_path, self.log_format)
        ed = CopyEditor(api)
        ed.parsers = self._get_parsers()
        ed.add_off_instruction("off")
        if path := task.get("request"):
            ed.set_instruction("on", path)
            ed.set_init_instruction("on")
        cmds = self._react_as_commands(task.get("react"))
        return Task(ed, cmds)

    def help(self) -> str:
        buf = io.StringIO()
        buf.write("task choices:\n")
        for name in self.task_names:
            buf.write("  ")
            buf.write(name)
            buf.write("\n")
            task = self._tasks[name]
            if path := task.get("request"):
                buf.write("    Request: ")
                buf.write(str(path))
                buf.write("\n")
            react = self._react_as_commands(task.get("react"))
            if react:
                buf.write("    React commands:\n")
                for r in react:
                    buf.write("      ")
                    buf.write(help_example_react(r))
            buf.write("\n")
        return buf.getvalue()


def help_example_react(cmd: str) -> str:
    subs = {
        '"$0"': "<source>",
        '"$1"': "<rev1>",
        '"$@"': "<rev1> ... <revN>",
    }
    for k, v in subs.items():
        cmd = cmd.replace(k, v)
    return cmd + "\n"



# Python standard libraries
import os, shutil
from importlib import resources
from pathlib import Path
from typing import Any

STD_BASE_DIRS = dict(
    TMPDIR="/tmp",
    XDG_CONFIG_HOME="~/.config",
    XDG_STATE_HOME="~/.local/state",
)


def get_std_path(env_var_name: str, subpath: str) -> Path:
    base_dir = os.environ.get(env_var_name)
    if base_dir is None:
        base_dir = STD_BASE_DIRS[env_var_name]
    return Path(base_dir).expanduser() / subpath


def copy_package_dir(package_path: str, dest_dir: Path) -> None:
    os.makedirs(dest_dir, exist_ok=True)
    quasidir = resources.files(__package__).joinpath(package_path)
    for quasipath in quasidir.iterdir():
        with resources.as_file(quasipath) as filepath:
            shutil.copy(filepath, dest_dir)


def resolve_path(ref_dir: Path, path: Any) -> Path | None:
    ret = None
    if path is not None:
        ret = Path(str(path)).expanduser()
        if not ret.is_absolute():
            ret = ref_dir / ret
    return ret


def read_file_text(file_path: Path | None) -> str | None:
    ret = None
    if file_path is not None:
        with open(file_path, 'r') as file:
            ret = file.read().strip()
    return ret



from .core import error, WorkFiles
from .task import Config, Task
from .util import get_std_path

# Python standard libraries
import argparse, logging
from pathlib import Path
from typing import Iterable

PROGNAME = "copyaid"
COPYAID_TMP_DIR = ("TMPDIR", "copyaid")
COPYAID_CONFIG_FILENAME = "copyaid.toml"
COPYAID_CONFIG_FILE = ("XDG_CONFIG_HOME", "copyaid/" + COPYAID_CONFIG_FILENAME)
COPYAID_LOG_DIR = ("XDG_STATE_HOME", "copyaid/log")
MAX_NUM_REVS = 7


def get_config_path(cmd_line_args: list[str] | None) -> Path | None:
    preparser = argparse.ArgumentParser(add_help=False)
    preparser.add_argument("-c", "--config", type=Path)
    (args, rest) = preparser.parse_known_args(cmd_line_args)
    if args.config and not args.config.exists():
        error(f"Config file '{args.config}' not found.")
        return None
    ret = args.config or Path(get_std_path(*COPYAID_CONFIG_FILE))
    if ret.is_dir():
        ret = ret / COPYAID_CONFIG_FILENAME
    return ret


def postconfig_argparser(
    task_names: Iterable[str], help_text: str
) -> argparse.ArgumentParser:
    parser = argparse.ArgumentParser(
        prog=PROGNAME,
        description="CopyAId",
        epilog=help_text,
        formatter_class=argparse.RawDescriptionHelpFormatter,
    )
    parser.add_argument(
        "-c",
        "--config",
        default=argparse.SUPPRESS,
        metavar="<config>",
        help="Configuration file"
    )
    parser.add_argument(
        "-d",
        "--dest",
        type=Path,
        metavar="<dest>",
        help="Destination directory for revisions"
    )
    parser.add_argument("task", choices=task_names, metavar="<task>")
    parser.add_argument("source", type=Path, nargs="+", metavar="<source>")
    return parser


def main(cmd_line_args: list[str] | None = None) -> int:
    logging.basicConfig()
    tmp_dir = Path(get_std_path(*COPYAID_TMP_DIR))
    config_path = get_config_path(cmd_line_args)
    if not config_path:
        return 2
    default_config = tmp_dir / "package"
    config = Config(default_config , config_path)
    help_text = (
        f"Default config: {default_config}/copyaid.toml\n" +
        f"User config: {config_path}\n\n" +
        config.help()
    )
    parser = postconfig_argparser(config.task_names, help_text)
    args = parser.parse_args(cmd_line_args)
    if args.dest is None:
        args.dest = tmp_dir
    exit_code = check_filename_collision(args.source)
    if exit_code != 0:
        return exit_code
    task = config.get_task(args.task, get_std_path(*COPYAID_LOG_DIR))
    for src in args.source:
        if not src.exists():
            error(f"File not found: '{src}'")
            exit_code = 2
            break
        work = WorkFiles(src, str(args.dest) + "/R{}/" + src.name, MAX_NUM_REVS)
        exit_code |= do_work(task, work)
        if exit_code > 1:
            break
    return exit_code


def check_filename_collision(sources: list[Path]) -> int:
    filenames = set()
    for s in sources:
        if s.name in filenames:
            msg = "Sources must have unique filenames. Conflict: {}"
            error(msg.format(s.name))
            return 2
        filenames.add(s.name)
    return 0


def do_work(task: Task, work: WorkFiles) -> int:
    if task.can_request:
        print("Saving revisions to", work.dest_glob)
        print(" for source", work.src)
        task.request(work)
    return task.react(work)



from copyaid.diff import diffadapt
import tomli

# Python Standard Library
import filecmp, json, logging, os
from dataclasses import dataclass
from datetime import datetime
from pathlib import Path
from typing import Any, Optional, TextIO
from typing_extensions import Protocol

LOGGER = logging.getLogger('copyaid')
error = LOGGER.error
warning = LOGGER.warning
info = LOGGER.info
debug = LOGGER.debug


class LiveOpenAiApi:
    def __init__(self, api_key: Optional[str] = None):
        from openai import OpenAI  # delay a slow import

        self.client = OpenAI(api_key=api_key)

    def query(self, req: Any) -> Any:
        return self.client.chat.completions.create(**req)


class PromptSettings:
    def __init__(self, path: Path):
        with open(path, "rb") as file:
            data = tomli.load(file)
        self.max_tokens_ratio = data["max_tokens_ratio"]
        self.system_prompt = data["chat_system"]
        self._openai = data.get("openai")
        self._prepend = data.get("prepend", "")
        self._append = data.get("append", "")

    @property
    def num_revisions(self) -> int:
        return int(self._openai.get("n", 1)) if self._openai else 1

    def make_openai_request(self, source: str) -> dict[str, Any]:
        assert isinstance(self._openai, dict)
        ret = dict(self._openai)
        ret["max_tokens"] = max(32, int(self.max_tokens_ratio * len(source) / 4))
        ret["messages"] = [
            {
                "role": "system",
                "content": self.system_prompt
            },
            {
                "role": "user",
                "content": self._prepend + source + self._append
            },
        ]
        return ret


class ApiProxy:
    ApiClass = LiveOpenAiApi

    def __init__(
        self, api_key: Optional[str], log_path: Path, log_format: Optional[str]
    ):
        self.log_path = log_path
        self.log_format = log_format
        self._api = ApiProxy.ApiClass(api_key)

    def do_request(self, settings: PromptSettings, text: str, name: str) -> list[str]:
        request = settings.make_openai_request(text)
        response = self._api.query(request)
        self.log_openai_query(name, request, response)
        return [c.message.content for c in response.choices]

    def log_openai_query(self, name: str, request: Any, response: Any) -> None:
        if not self.log_format:
            return
        t = datetime.utcfromtimestamp(response.created)
        ts = t.isoformat().replace("-", "").replace(":", "") + "Z"
        response_ex = {
            'choices': {
                '__all__': {
                    'logprobs': {
                        'content': {
                            '__all__': {
                                'bytes': True,
                                'top_logprobs': {
                                    0: True,
                                    '__all__': {'bytes'},
                                },
                            }
                        },
                    },
                }
            },
        }
        response_dump = response.model_dump(exclude_unset=True, exclude=response_ex)
        data = dict(request=request, response=response_dump)
        os.makedirs(self.log_path, exist_ok=True)
        save_stem = name + "." + ts
        print("Logging OpenAI response", save_stem)
        if self.log_format == "jsoml":
            import jsoml
            jsoml.dump(data, self.log_path / (save_stem + ".xml"))
        elif self.log_format == "json":
            with open(self.log_path / (save_stem + ".json"), "w") as file:
                json.dump(data, file, indent=4, ensure_ascii=False)
                file.write("\n")
        else:
            warning("Unsupported log format: {}".format(self.log_format))


class WorkFiles:
    def __init__(self, src: str | Path, dest: str | Path, max_num_revs: int = 1):
        assert 0 < max_num_revs < 10
        self.src = Path(src)
        dest = str(dest)
        self._dests = [Path(dest.format(i + 1)) for i in range(max_num_revs)]
        self.dest_glob = dest.format("?")
        self._files: list[TextIO] = list()

    def revisions(self) -> list[Path]:
        return [p for p in self._dests if p.exists()]

    def open_new_dests(self, n: int = 1) -> None:
        assert n > 0
        self._files = []
        for i, path in enumerate(self._dests):
            if i < n:
                os.makedirs(path.parent, exist_ok=True)
                self._files.append(open(path, "w"))
            else:
                path.unlink(missing_ok=True)

    def write_dest(self, text: str, i: int = 1) -> None:
        self._files[i].write(text)
        self._files[i].flush()

    def close_dests(self) -> None:
        for f in self._files:
            f.close()
        self._files = []


@dataclass
class Copybreak:
    raw_line: str
    args: list[str]

    @property
    def keyword(self) -> str | None:
        return self.args[0] if self.args else None

    @property
    def instruction(self) -> str | None:
        return self.args[1] if len(self.args) > 1 else None


@dataclass
class TextSegment:
    copybreak: Copybreak | None
    text: str


class ParsedSource:
    def __init__(self) -> None:
        self.segments: list[TextSegment] = list()

    def instructions(self) -> set[str]:
        ret = set()
        for seg in self.segments:
            if seg.copybreak and seg.copybreak.instruction:
                ret.add(seg.copybreak.instruction)
        return ret


class SourceParserProtocol(Protocol):
    def parse(self, src: Path) -> ParsedSource | None:
        ...


def parse_source(parsers: list[SourceParserProtocol], src: Path) -> ParsedSource:
    for parser in parsers:
        if ret := parser.parse(src):
            return ret
    raise RuntimeError(f"Not able to parse format of {src}")


class TrivialParser:
    def parse(self, src_path: Path) -> ParsedSource | None:
        ret = ParsedSource()
        warning(f"No file format configured for: {src_path}")
        with open(src_path) as file:
            ret.segments.append(TextSegment(None, file.read()))
        return ret


@dataclass
class CopybreakSyntax:
    keywords: list[str]
    prefix: str
    suffix: str | None

    def parse(self, line: str) -> Copybreak | None:
        s = line.strip()
        if not s.startswith(self.prefix):
            return None
        s = s[len(self.prefix):]
        idx = 0
        if self.suffix:
            idx = s.find(self.suffix)
            if idx > 0:
                s = s[:idx]
        candidate = Copybreak(line, s.split())
        if candidate.keyword not in self.keywords:
            return None
        if idx < 0:
            warning("Copybreak line missing suffix '{}'".format(self.suffix))
        return candidate

    @staticmethod
    def from_POD(pod: dict[str, Any]) -> "CopybreakSyntax":
        return CopybreakSyntax(pod["keywords"], pod["prefix"], pod.get("suffix"))


class SimpleParser:
    def __init__(self, copybreak: CopybreakSyntax):
        self.copybreak = copybreak
        self.extensions_filter: list[str] | None = None

    def parse(self, src: Path) -> ParsedSource | None:
        if self.extensions_filter is not None:
            if src.suffix not in self.extensions_filter:
                return None
        ret = ParsedSource()
        pending_copybreak = None
        pending_lines: list[str] = list()
        with open(src) as file:
            for line in file:
                if new_copybreak := self.copybreak.parse(line):
                    segment = TextSegment(pending_copybreak, "".join(pending_lines))
                    ret.segments.append(segment)
                    pending_copybreak = new_copybreak
                    pending_lines = list()
                else:
                    pending_lines.append(line)
            segment = TextSegment(pending_copybreak, "".join(pending_lines))
            ret.segments.append(segment)
        return ret

    @staticmethod
    def from_POD(pod: dict[str, Any]) -> "SimpleParser":
        ret = SimpleParser(CopybreakSyntax.from_POD(pod["copybreak"]))
        ret.extensions_filter = pod.get("extensions")
        return ret


class CopyEditor:
    def __init__(self, api: ApiProxy):
        self.api = api
        self.parsers: list[SourceParserProtocol] = []
        self._instructions: dict[str, PromptSettings | None] = dict()

    @property
    def has_instructions(self) -> bool:
        return any(bool(p) for p in self._instructions.values())

    def set_instruction(self, instruction: str, settings: Path | str) -> None:
        self._instructions[instruction] = PromptSettings(Path(settings))

    def add_off_instruction(self, instruction: str) -> None:
        self._instructions[instruction] = None

    def set_init_instruction(self, instruction: str | None) -> None:
        settings = self._instructions[instruction] if instruction else None
        self._instructions[""] = settings

    def _num_revisions(self, src: ParsedSource) -> int:
        ret = 1
        if init_instr := self._instructions.get(""):
            ret = init_instr.num_revisions
        for iid in src.instructions():
            if iid not in self._instructions:
                raise SyntaxError(f"'{iid}' is not a configured copybreak instruction.")
            if instr := self._instructions.get(iid):
                assert instr.num_revisions > 0
                if ret == 1:
                    if instr.num_revisions > ret:
                        ret = instr.num_revisions
                elif instr.num_revisions > 1:
                    if instr.num_revisions < ret:
                        ret = instr.num_revisions
                        warning(f"Instruction {iid} sets number of revisions to {ret}.")
                    elif instr.num_revisions > ret:
                        msg = "Only {} of {} revisions used with instruction {}."
                        warning(msg.format(ret, instr.num_revisions, iid))
        return ret

    def revise(self, work: WorkFiles) -> None:
        parsed = parse_source(self.parsers, work.src)
        num_revisions = self._num_revisions(parsed)
        work.open_new_dests(num_revisions)
        cur_settings = self._instructions.get("")
        for si, seg in enumerate(parsed.segments):
            if seg.copybreak:
                for ri in range(num_revisions):
                    work.write_dest(seg.copybreak.raw_line, ri)
                if seg.copybreak.instruction:
                    cur_settings = self._instructions[seg.copybreak.instruction]
            log_name = "{}.{}".format(work.src.stem, si)
            if cur_settings and len(seg.text.strip()):
                revisions = self.api.do_request(cur_settings, seg.text, log_name)
                if len(revisions) > num_revisions:
                    revisions = revisions[:num_revisions]
                elif len(revisions) == 1 and num_revisions > 1:
                    revisions = list(revisions[0]) * num_revisions
                assert len(revisions) == num_revisions
                revisions = diffadapt(seg.text, revisions)
            else:
                revisions = [seg.text] * num_revisions
            for ri, rev in enumerate(revisions):
                work.write_dest(rev, ri)
        work.close_dests()



import pytest

import copyaid.cli

import os
from types import SimpleNamespace

SOURCE_TEXT = "Jupiter big.\nJupiter a planet.\nJupiter gas.\n"
MOCK_COMPLETION = "Jupiter is a big planet made of gas."
EXPECTED_TEXT = "Jupiter is\na big planet made of\ngas.\n"

class MockApi:
    def __init__(self, api_key):
        pass

    def query(self, req):
        return SimpleNamespace(
            created=1674259148,
            choices=[
                SimpleNamespace(
                    message=SimpleNamespace(
                        content=MOCK_COMPLETION
                    )
                )
            ]
        )

copyaid.core.ApiProxy.ApiClass = MockApi


def get_revision(src_path, src_text, task="proof"):
    open(src_path, "w").write(src_text)
    retcode = copyaid.cli.main([
        task,
        str(src_path),
        "--dest", str(src_path.parent),
        "--config", "tests/mock_config.toml",
    ])
    assert retcode == 0
    return open(src_path.parent / "R1" / src_path.name).read()


def test_main(tmp_path):
    got = get_revision(tmp_path / "source.txt", SOURCE_TEXT)
    assert got == EXPECTED_TEXT


def test_copybreak_md(tmp_path):
    copybreak = "<!-- copybreak -->\n"
    src_text = SOURCE_TEXT + copybreak + SOURCE_TEXT
    got = get_revision(tmp_path / "source.md", src_text)
    assert got == EXPECTED_TEXT + copybreak + EXPECTED_TEXT

def test_copybreak_off_md(tmp_path):
    copybreak = "<!-- copybreak off -->\n"
    src_text = SOURCE_TEXT + copybreak + SOURCE_TEXT
    got = get_revision(tmp_path / "source.md", src_text)
    assert got == EXPECTED_TEXT + copybreak + SOURCE_TEXT

def test_copybreak_tex(tmp_path):
    copybreak = "%% copybreak\n"
    src_text = SOURCE_TEXT + copybreak + SOURCE_TEXT
    got = get_revision(tmp_path / "source.tex", src_text)
    assert got == EXPECTED_TEXT + copybreak + EXPECTED_TEXT

def test_copybreak_off_tex(tmp_path):
    copybreak = "%% copybreak off\n"
    src_text = SOURCE_TEXT + copybreak + SOURCE_TEXT
    got = get_revision(tmp_path / "source.tex", src_text)
    assert got == EXPECTED_TEXT + copybreak + SOURCE_TEXT

def test_copybreak_foobar(tmp_path):
    copybreak = "¡¿ copybreak ?!\n"
    src_text = SOURCE_TEXT + copybreak + SOURCE_TEXT
    got = get_revision(tmp_path / "source.foobar", src_text, "fooit")
    assert got == EXPECTED_TEXT + copybreak + EXPECTED_TEXT

def test_copybreak_off_foobar(tmp_path):
    copybreak = "¡¿ copybreak off ?!\n"
    src_text = SOURCE_TEXT + copybreak + SOURCE_TEXT
    got = get_revision(tmp_path / "source.foobar", src_text, "fooit")
    assert got == EXPECTED_TEXT + copybreak + SOURCE_TEXT



# Python Standard Library
import difflib, re

from typing import Iterator, Optional

###############################################################################
# Code that "diff-adapts" altered text to be more diff-friendly with original
###############################################################################

Tokens = list[str]


class TokenSequenceMatcher:

    # just an opaque token object to represent end-of-message for matching algo
    EOM = "just-random-a0f75a980e88b9c27fa02ed5b8def537d131f281"

    def __init__(self, focal_text: str):
        isjunk = lambda x: x == " "
        self.matcher = difflib.SequenceMatcher(isjunk, autojunk=False)
        self.re_token = re.compile(r"\w+|\W|\n")
        self.focus = self.tokenize(focal_text) + [TokenSequenceMatcher.EOM]
        # SequenceMatcher: ... caches detailed information about the second sequence,
        # so if you want to compare one sequence against many sequences,
        # use set_seq2() ...
        self.matcher.set_seq2(self.focus)

    def tokenize(self, text: str) -> Tokens:
        return [match[0] for match in self.re_token.finditer(text)]

    def set_alternative(self, alt_text: str) -> None:
        self.alt = self.tokenize(alt_text) + [TokenSequenceMatcher.EOM]
        self.matcher.set_seq1(self.alt)

    def operations(self) -> Iterator[tuple[str, Tokens, Tokens]]:
        """tag meaning is relative to going from alt text to focal text"""

        return (
            (tag, self.alt[a1:a2], self.focus[f1:f2])
            for tag, a1, a2, f1, f2 in self.matcher.get_opcodes()
        )


class DiffAdaptor:
    def __init__(self) -> None:
        self.last_token: Optional[str] = None
        self.line_debt = 0

    @staticmethod
    def apply_operations(matcher: TokenSequenceMatcher) -> str:
        tokens = []
        adaptor = DiffAdaptor()
        # ops for converting revised text back to orig
        for tag, rev_chunk, orig_chunk in matcher.operations():
            tokens += adaptor._do_operation(tag, rev_chunk, orig_chunk)
        assert adaptor.last_token == TokenSequenceMatcher.EOM
        return "".join(tokens)

    def _do_operation(self, tag: str, rev: Tokens, orig: Tokens) -> Tokens:
        assert rev or orig
        ret = []
        if tag == "equal":
            ret = self._adapt_unrevised(rev)
        else:
            self.line_debt += orig.count("\n")
            if len(rev) > 0:
                ret = self._adapt_revised(rev, orig)
            else:
                if self._undo_delete(orig):
                    self.line_debt -= orig.count("\n")
                    ret = orig
        if ret:
            if self.last_token is not None:
                ret.insert(0, self.last_token)
            self.last_token = ret.pop()
        return ret

    def _undo_delete(self, orig: Tokens) -> bool:
        assert orig
        new_line = (orig[0] == "\n" or self.last_token in ("\n", None))
        if new_line and all(s.isspace() for s in orig):
            # undo deletion of indentation and extra newlines
            return True
        return False

    def _adapt_unrevised(self, chunk: Tokens) -> Tokens:
        assert chunk
        # a large line debt is suspect, probably bad sequence matching
        if self.line_debt == 1:
            if self.last_token == " ":
                self.last_token = "\n"
                self.line_debt -= 1
            elif self.last_token != "\n":
                if chunk[0] == " ":
                    chunk[0] = "\n"
                    self.line_debt -= 1
        if len(chunk) > 2:
            # a non-trivial unrevised chunk probably means lines are now aligning
            self.line_debt = 0
        return chunk

    def _adapt_revised(self, rev: Tokens, orig: Tokens) -> Tokens:
        assert rev or orig
        if orig in (['\n'], ['.', '\n'], [',', '\n'], [';', '\n']):
            if rev[0] == " ":
                rev[0] = "\n"
            elif rev[0] in (".", ",", ";"):
                if rev[1:] == [" "]:
                    rev[1] = "\n"
                else:
                    rev.insert(1, "\n")
            else:
                rev.insert(0, "\n")
        prev_token = self.last_token
        for i in range(len(rev)):
            if rev[i] == "\n":
                self.line_debt -= 1
            elif self.line_debt > 0:
                if prev_token in (".", ",", ";"):
                    if rev[i] == " ":
                        rev[i] = "\n"
                        self.line_debt -= 1
            prev_token = rev[i]
        if not orig and rev.count("\n") > 1:
            # insertion of multiple lines
            self.line_debt = 0
        return rev


def diffadapt(
    orig_text: str,
    revisions: list[str],
    codeword: Optional[str] = None,
) -> list[str]:
    ret = []
    matcher = TokenSequenceMatcher(orig_text)
    for rev_text in revisions:
        if codeword == rev_text.strip():
            ret.append(orig_text)
        else:
            if not rev_text.endswith("\n"):
                rev_text += "\n"
            matcher.set_alternative(rev_text)
            ret.append(DiffAdaptor.apply_operations(matcher))
    return ret


def cli(cmd_line_args: Optional[list[str]] = None) -> int:
    import argparse, pathlib
    parser = argparse.ArgumentParser(
        prog="diffadapt",
        description="Adapt revised files to be more diff-friendly with the source.",
    )
    parser.add_argument("src")
    parser.add_argument("rev", nargs="+")
    parser.add_argument("-c", "--codeword", help="Codeword for no changes.")
    args = parser.parse_args(cmd_line_args)
    with open(args.src) as file:
        source_text = file.read()
    rev_texts = list()
    for rev in args.rev:
        with open(rev) as file:
            rev_texts.append(file.read())
    rev_texts = diffadapt(source_text, rev_texts, args.codeword)
    for i, rev in enumerate(args.rev):
        with open(rev, "w") as file:
            file.write(rev_texts[i])
    return 0
